"""rio_tiler.reader.sentinel2: Sentinel-2 processing."""

from typing import Any, Dict, Sequence, Tuple, Union

import os
import re
import itertools
from collections import OrderedDict

import numpy

import rasterio
from rasterio.warp import transform_bounds

from rio_tiler import reader
from rio_tiler import constants
from rio_tiler.utils import tile_exists
from rio_tiler.errors import TileOutsideBounds, InvalidBandName, InvalidSentinelSceneId


SENTINEL_L1_BANDS = OrderedDict(
    [
        ("10", ["02", "03", "04", "08"]),
        ("20", ["05", "06", "07", "11", "12", "8A"]),
        ("60", ["01", "09", "10"]),
    ]
)

SENTINEL_L2_BANDS = OrderedDict(
    [
        ("10", ["02", "03", "04", "08"]),
        ("20", ["02", "03", "04", "05", "06", "07", "08", "11", "12", "8A"]),
        (
            "60",
            ["01", "02", "03", "04", "05", "06", "07", "08", "09", "11", "12", "8A"],
        ),
    ]
)

SENTINEL_L2_PRODUCTS = OrderedDict(
    [
        ("10", ["AOT", "WVP"]),
        ("20", ["AOT", "SCL", "WVP"]),
        ("60", ["AOT", "SCL", "WVP"]),
    ]
)


def sentinel2_parser(sceneid: str) -> Dict:
    """
    Parse Sentinel-2 scene id.

    Attributes
    ----------
        sceneid : str
            Sentinel-2 sceneid.

    Returns
    -------
        out : dict
            dictionary with metadata constructed from the sceneid.

    """

    if not re.match("^S2[AB]_L[0-2][A-C]_[0-9]{8}_[0-9]{2}[A-Z]{3}_[0-9]$", sceneid):
        raise InvalidSentinelSceneId("Could not match {}".format(sceneid))

    sentinel_pattern = (
        r"^S"
        r"(?P<sensor>\w{1})"
        r"(?P<satellite>[AB]{1})"
        r"_"
        r"(?P<processingLevel>L[0-2][ABC])"
        r"_"
        r"(?P<acquisitionYear>[0-9]{4})"
        r"(?P<acquisitionMonth>[0-9]{2})"
        r"(?P<acquisitionDay>[0-9]{2})"
        r"_"
        r"(?P<utm>[0-9]{2})"
        r"(?P<lat>\w{1})"
        r"(?P<sq>\w{2})"
        r"_"
        r"(?P<num>[0-9]{1})$"
    )

    meta: Dict[str, Any] = re.match(
        sentinel_pattern, sceneid, re.IGNORECASE
    ).groupdict()
    meta["scene"] = sceneid

    utm_zone = meta["utm"].lstrip("0")
    grid_square = meta["sq"]
    latitude_band = meta["lat"]
    year = meta["acquisitionYear"]
    month = meta["acquisitionMonth"].lstrip("0")
    day = meta["acquisitionDay"].lstrip("0")
    img_num = meta["num"]

    meta["scheme"] = "s3"
    meta["bucket"] = "sentinel-s2-" + meta["processingLevel"].lower()
    meta["prefix"] = os.path.join(
        "tiles", utm_zone, latitude_band, grid_square, year, month, day, img_num
    )

    if meta["processingLevel"] == "L1C":
        meta["preview_file"] = "preview.jp2"
        meta["preview_prefix"] = ""
        meta["bands"] = list(
            itertools.chain.from_iterable(
                [bands for _, bands in SENTINEL_L1_BANDS.items()]
            )
        )
        meta["valid_bands"] = meta["bands"]
    else:
        meta["preview_file"] = "R60m/TCI.jp2"
        meta["preview_prefix"] = "R60m"
        meta["bands"] = SENTINEL_L2_BANDS["60"]
        meta["valid_bands"] = meta["bands"] + SENTINEL_L2_PRODUCTS["60"]

    return meta


def _l2_prefixed_band(band: str) -> str:
    """Return L2A prefixed bands name."""
    if band in SENTINEL_L2_BANDS["60"]:
        for res, bands in SENTINEL_L2_BANDS.items():
            if band in bands:
                return "R{}m/B{}".format(res, band)
    elif band in SENTINEL_L2_PRODUCTS["60"]:
        for res, bands in SENTINEL_L2_PRODUCTS.items():
            if band in bands:
                return "R{}m/{}".format(res, band)

    raise InvalidBandName("{} is not a valid Sentinel band name".format(band))


def bounds(sceneid: str) -> Dict:
    """
    Retrieve image bounds.

    Attributes
    ----------
        sceneid : str
            Sentinel-2 sceneid.

    Returns
    -------
        out : dict
            dictionary with image bounds.

    """
    scene_params = sentinel2_parser(sceneid)
    preview_file = "{scheme}://{bucket}/{prefix}/{preview_file}".format(**scene_params)
    with rasterio.open(preview_file) as src_dst:
        bounds = transform_bounds(
            src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
        )

    return dict(sceneid=sceneid, bounds=bounds)


def metadata(
    sceneid: str,
    pmin: float = 2.0,
    pmax: float = 98.0,
    hist_options: Dict = {},
    **kwargs: Any,
) -> Dict:
    """
    Retrieve image bounds and band statistics.

    Attributes
    ----------
        sceneid : str
            Sentinel-2 sceneid.
        pmin : float, optional, (default: 2.)
            Histogram minimum cut.
        pmax : float, optional, (default: 98.)
            Histogram maximum cut.
        hist_options : dict, optional
            Options to forward to numpy.histogram function.
            e.g: {bins=20, range=(0, 1000)}
        kwargs : optional
            These are passed to 'rio_tiler.reader.preview'

    Returns
    -------
        out : dict
            Dictionary with image bounds and bands statistics.

    """
    scene_params = sentinel2_parser(sceneid)
    sentinel_prefix = "{scheme}://{bucket}/{prefix}/{preview_prefix}".format(
        **scene_params
    )
    bands = scene_params["bands"]

    addresses = [f"{sentinel_prefix}/B{band}.jp2" for band in bands]

    responses = reader.multi_metadata(
        addresses,
        indexes=[1],
        nodata=0,
        percentiles=(pmin, pmax),
        hist_options=hist_options,
        **kwargs,
    )
    info: Dict[str, Any] = dict(sceneid=sceneid)
    info["band_descriptions"] = [(ix + 1, b) for ix, b in enumerate(bands)]
    info["bounds"] = responses[0]["bounds"]
    info["statistics"] = {
        b: v for b, d in zip(bands, responses) for k, v in d["statistics"].items()
    }
    return info


def tile(
    sceneid: str,
    tile_x: int,
    tile_y: int,
    tile_z: int,
    bands: Union[Sequence[str], str] = ("04", "03", "02"),
    tilesize: int = 256,
    **kwargs: Dict,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Create mercator tile from Sentinel-2 data.

    Attributes
    ----------
        sceneid : str
            Sentinel-2 sceneid.
        tile_x : int
            Mercator tile X index.
        tile_y : int
            Mercator tile Y index.
        tile_z : int
            Mercator tile ZOOM level.
        bands : tuple, str, optional (default: ('04', '03', '02'))
            Bands index for the RGB combination.
        tilesize : int, optional (default: 256)
            Output image size.
        kwargs: dict, optional
            These will be passed to the 'rio_tiler.utils._tile_read' function.

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    if isinstance(bands, str):
        bands = (bands,)

    scene_params = sentinel2_parser(sceneid)
    for band in bands:
        if band not in scene_params["valid_bands"]:
            raise InvalidBandName("{} is not a valid Sentinel band name".format(band))

    sentinel_prefix = "{scheme}://{bucket}/{prefix}".format(**scene_params)

    preview_file = os.path.join(sentinel_prefix, scene_params["preview_file"])
    with rasterio.open(preview_file) as src:
        bounds = transform_bounds(
            src.crs, constants.WGS84_CRS, *src.bounds, densify_pts=21
        )
        if not tile_exists(bounds, tile_z, tile_x, tile_y):
            raise TileOutsideBounds(
                "Tile {}/{}/{} is outside image bounds".format(tile_z, tile_x, tile_y)
            )

    if scene_params["processingLevel"] == "L2A":
        bands = [_l2_prefixed_band(b) for b in bands]
    else:
        bands = ["B{}".format(b) for b in bands]

    addresses = [f"{sentinel_prefix}/{band}.jp2" for band in bands]
    return reader.multi_tile(
        addresses, tile_x, tile_y, tile_z, tilesize=tilesize, nodata=0
    )



"""rio-tiler.reader: image utility functions."""

from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import math
import warnings
from concurrent import futures

import numpy

from affine import Affine

import mercantile

import rasterio
from rasterio.io import DatasetReader, DatasetWriter
from rasterio.crs import CRS
from rasterio.vrt import WarpedVRT
from rasterio.warp import transform_bounds, transform
from rasterio.enums import Resampling, ColorInterp
from rasterio.windows import Window

from rio_tiler import constants
from rio_tiler.utils import (
    get_vrt_transform,
    has_alpha_band,
    has_mask_band,
    non_alpha_indexes,
    _stats as raster_stats,
    _requested_tile_aligned_with_internal_tile as is_aligned,
    tile_exists,
)
from rio_tiler.errors import TileOutsideBounds, AlphaBandWarning


def _read(
    src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT],
    height: int,
    width: int,
    indexes: Optional[Union[Sequence[int], int]] = None,
    out_window: Optional[Window] = None,
    nodata: Optional[Union[float, int, str]] = None,
    resampling_method: Resampling = "bilinear",
    force_binary_mask: bool = True,
    unscale: bool = False,
    vrt_options: Dict = {},
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Create WarpedVRT and read data and mask.

    Attributes
    ----------
        src_dst : rasterio.io.DatasetReader
            rasterio.io.DatasetReader object
        height : int
            Output height of the array.
        width : int
            Output width of the array.
        indexes : list of ints or a single int, optional
            Band indexes
        out_window: rasterio.windows.Window, optional
            Output window to read.
        nodata: int or float, optional
        resampling_method : str, optional
            Resampling algorithm. Default is "bilinear".
        force_binary_mask, bool, optional
            If True, rio-tiler makes sure mask has only 0 or 255 values.
            Default is set to True.
        unscale, bool, optional
            If True, apply scale and offset to the data array.
            Default is set to False.
        vrt_options: dict, optional
            These will be passed to the rasterio.warp.WarpedVRT class.

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    if isinstance(indexes, int):
        indexes = (indexes,)

    vrt_params = dict(add_alpha=True, resampling=Resampling[resampling_method])
    nodata = nodata if nodata is not None else src_dst.nodata
    if nodata is not None:
        vrt_params.update(dict(nodata=nodata, add_alpha=False, src_nodata=nodata))

    if has_alpha_band(src_dst):
        vrt_params.update(dict(add_alpha=False))

    if indexes is None:
        indexes = non_alpha_indexes(src_dst)
        if indexes != src_dst.indexes:
            warnings.warn(
                "Alpha band was removed from the output data array", AlphaBandWarning
            )

    vrt_params.update(vrt_options)
    with WarpedVRT(src_dst, **vrt_params) as vrt:
        data = vrt.read(
            out_shape=(len(indexes), height, width),
            indexes=indexes,
            window=out_window,
            resampling=Resampling[resampling_method],
        )
        if ColorInterp.alpha in vrt.colorinterp:
            idx = vrt.colorinterp.index(ColorInterp.alpha) + 1
            mask = vrt.read(
                indexes=idx,
                out_shape=(height, width),
                window=out_window,
                resampling=Resampling[resampling_method],
                out_dtype="uint8",
            )
        else:
            mask = vrt.dataset_mask(
                out_shape=(height, width),
                window=out_window,
                resampling=Resampling[resampling_method],
            )

        if force_binary_mask:
            mask = numpy.where(mask != 0, numpy.uint8(255), numpy.uint8(0))

        if unscale:
            data = data.astype("float32", casting="unsafe")
            numpy.multiply(data, vrt.scales[0], out=data, casting="unsafe")
            numpy.add(data, vrt.offsets[0], out=data, casting="unsafe")

    return data, mask


def part(
    src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT],
    bounds: Tuple[float, float, float, float],
    height: int,
    width: int,
    padding: int = 0,
    dst_crs: CRS = constants.WEB_MERCATOR_CRS,
    bounds_crs: Optional[CRS] = None,
    minimum_overlap: Optional[float] = None,
    warp_vrt_option: Dict = {},
    **kwargs: Any,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Read part of an image.

    Attributes
    ----------
        src_dst : rasterio.io.DatasetReader
            rasterio.io.DatasetReader object
        bounds : tuple
            Output bounds (left, bottom, right, top) in target crs ("dst_crs").
        height : int
            Output height of the array.
        width : int
            Output width of the array.
        padding : int, optional
            Padding to apply to each edge of the tile when retrieving data
            to assist in reducing resampling artefacts along edges.
        dst_crs: CRS or str, optional
            Target coordinate reference system, default is "epsg:3857".
        bounds_crs: CRS or str, optional
            Overwrite bounds coordinate reference system, default is equal
            to the output CRS (dst_crs).
        minimum_tile_cover: float, optional
            Minimum % overlap for which to raise an error with dataset not
            covering enought of the tile.
        warp_vrt_option: dict, optional
            These will be passed to the rasterio.warp.WarpedVRT class.
        kwargs : Any, optional
            Additional options to forward to reader._read()

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    if not bounds_crs:
        bounds_crs = dst_crs

    bounds = transform_bounds(bounds_crs, dst_crs, *bounds, densify_pts=21)

    if minimum_overlap:
        src_bounds = transform_bounds(
            src_dst.crs, dst_crs, *src_dst.bounds, densify_pts=21
        )
        x_overlap = max(
            0, min(src_bounds[2], bounds[2]) - max(src_bounds[0], bounds[0])
        )
        y_overlap = max(
            0, min(src_bounds[3], bounds[3]) - max(src_bounds[1], bounds[1])
        )
        cover_ratio = (x_overlap * y_overlap) / (
            (bounds[2] - bounds[0]) * (bounds[3] - bounds[1])
        )

        if cover_ratio < minimum_overlap:
            raise TileOutsideBounds(
                "Dataset covers less than {:.0f}% of tile".format(cover_ratio * 100)
            )

    vrt_transform, vrt_width, vrt_height = get_vrt_transform(
        src_dst, bounds, width, height, dst_crs=dst_crs
    )

    out_window = Window(col_off=0, row_off=0, width=vrt_width, height=vrt_height)

    if padding > 0 and not is_aligned(src_dst, bounds, height, width):
        vrt_transform = vrt_transform * Affine.translation(-padding, -padding)
        orig_vrt_height = vrt_height
        orig_vrt_width = vrt_width
        vrt_height = vrt_height + 2 * padding
        vrt_width = vrt_width + 2 * padding
        out_window = Window(
            col_off=padding,
            row_off=padding,
            width=orig_vrt_width,
            height=orig_vrt_height,
        )

    return _read(
        src_dst,
        height,
        width,
        out_window=out_window,
        vrt_options=dict(
            crs=dst_crs,
            transform=vrt_transform,
            width=vrt_width,
            height=vrt_height,
            **warp_vrt_option,
        ),
        **kwargs,
    )


def preview(
    src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT],
    max_size: int = 1024,
    height: int = None,
    width: int = None,
    **kwargs: Any,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Read image and resample to low resolution.


    Attributes
    ----------
        src_dst : rasterio.io.DatasetReader
            rasterio.io.DatasetReader object
        max_size : int
            `max_size` of the longest dimension, respecting
            bounds X/Y aspect ratio.
        height: int, optional
            output height of the data
        width: int, optional
            output width of the data
        kwargs : Any, optional
            Additional options to forward to reader._read()

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    if not height and not width:
        if max(src_dst.height, src_dst.width) < max_size:
            height, width = src_dst.height, src_dst.width
        else:
            ratio = src_dst.height / src_dst.width
            if ratio > 1:
                height = max_size
                width = math.ceil(height / ratio)
            else:
                width = max_size
                height = math.ceil(width * ratio)

    return _read(src_dst, height, width, **kwargs)


def point(
    src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT],
    coordinates: Tuple[float, float],
    indexes: Optional[Union[Sequence[int], int]] = None,
    coord_crs: CRS = constants.WGS84_CRS,
    unscale: bool = False,
) -> List:
    """
    Read point value

    Attributes
    ----------
        src_dst : rasterio.io.DatasetReader
            rasterio.io.DatasetReader object
        coordinates : tuple
            (X, Y) coordinates.
        indexes : list of ints or a single int, optional
            Band indexes
        coord_crs : rasterio.crs.CRS, optional
            (X, Y) coordinate system. Default is WGS84/EPSG:4326.
        unscale, bool, optional
            If True, apply scale and offset to the data.
            Default is set to False.
        kwargs : Any, optional
            Additional options to forward to src_dst.sample()

    Returns
    -------
        point : list
            List of pixel values per bands indexes.

    """
    if isinstance(indexes, int):
        indexes = (indexes,)

    lon, lat = transform(coord_crs, src_dst.crs, [coordinates[0]], [coordinates[1]])
    if not (
        (src_dst.bounds[0] < lon[0] < src_dst.bounds[2])
        and (src_dst.bounds[1] < lat[0] < src_dst.bounds[3])
    ):
        raise Exception("Point is outside dataset bounds")

    indexes = indexes if indexes is not None else src_dst.indexes

    point_values = list(src_dst.sample([(lon[0], lat[0])], indexes=indexes))[0]

    if unscale:
        point_values = point_values.astype("float32", casting="unsafe")
        numpy.multiply(
            point_values, src_dst.scales[0], out=point_values, casting="unsafe"
        )
        numpy.add(point_values, src_dst.offsets[0], out=point_values, casting="unsafe")

    return point_values.tolist()


def metadata(
    src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT],
    bounds: Optional[Tuple[float, float, float, float]] = None,
    indexes: Optional[Union[Sequence[int], int]] = None,
    max_size: int = 1024,
    bounds_crs: CRS = constants.WGS84_CRS,
    percentiles: Tuple[float, float] = (2.0, 98.0),
    hist_options: Dict = {},
    **kwargs: Any,
) -> Dict:
    """
    Retrieve statistics from multiple sentinel bands.

    Attributes
    ----------
        src_dst : rasterio.io.DatasetReader
            rasterio.io.DatasetReader object
        bounds : tuple, optional
            Bounding box coordinates from which to calculate image statistics.
        max_size : int
            `max_size` of the longest dimension, respecting
            bounds X/Y aspect ratio.
        indexes : list of ints or a single int, optional
            Band indexes.
        bounds_crs: CRS or str, optional
            Specify bounds coordinate reference system, default WGS84/EPSG4326.
        percentiles: tuple, optional
            Tuple of Min/Max percentiles to compute. Default is (2, 98).
        hist_options : dict, optional
            Options to forward to numpy.histogram function.
        kwargs : Any, optional
            Additional options to forward to part or preview

    Returns
    -------
        dict

    """
    if isinstance(indexes, int):
        indexes = (indexes,)

    if indexes is None:
        indexes = non_alpha_indexes(src_dst)
        if indexes != src_dst.indexes:
            warnings.warn(
                "Alpha band was removed from the output data array", AlphaBandWarning
            )

    if bounds:
        asp_ratio = abs((bounds[2] - bounds[0]) / (bounds[1] - bounds[3]))
        if asp_ratio > 1:
            width = max_size
            height = math.ceil(max_size / asp_ratio)
        else:
            width = math.ceil(max_size * asp_ratio)
            height = max_size

        data, mask = part(
            src_dst,
            bounds,
            height,
            width,
            indexes=indexes,
            dst_crs=src_dst.crs,
            bounds_crs=bounds_crs,
            **kwargs,
        )
        bounds = transform_bounds(
            bounds_crs, constants.WGS84_CRS, *bounds, densify_pts=21
        )

    else:
        data, mask = preview(src_dst, max_size=max_size, indexes=indexes, **kwargs)
        bounds = transform_bounds(
            src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
        )

    data = numpy.ma.array(data)
    data.mask = mask == 0

    statistics = {
        indexes[b]: raster_stats(data[b], percentiles=percentiles, **hist_options)
        for b in range(data.shape[0])
    }

    def _get_descr(ix):
        """Return band description."""
        name = src_dst.descriptions[ix - 1]
        if not name:
            name = "band{}".format(ix)
        return name

    band_descriptions = [(ix, _get_descr(ix)) for ix in indexes]
    tags = [(ix, src_dst.tags(ix)) for ix in indexes]

    other_meta = dict()
    if src_dst.scales[0] and src_dst.offsets[0]:
        other_meta.update(dict(scale=src_dst.scales[0]))
        other_meta.update(dict(offset=src_dst.offsets[0]))

    if has_alpha_band(src_dst):
        nodata_type = "Alpha"
    elif has_mask_band(src_dst):
        nodata_type = "Mask"
    elif src_dst.nodata is not None:
        nodata_type = "Nodata"
    else:
        nodata_type = "None"

    try:
        cmap = src_dst.colormap(1)
        other_meta.update(dict(colormap=cmap))
    except ValueError:
        pass

    return dict(
        bounds=bounds,
        statistics=statistics,
        band_metadata=tags,
        band_descriptions=band_descriptions,
        dtype=src_dst.meta["dtype"],
        colorinterp=[src_dst.colorinterp[ix - 1].name for ix in indexes],
        nodata_type=nodata_type,
        **other_meta,
    )


def tile(
    src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT],
    x: int,
    y: int,
    z: int,
    tilesize: int = 256,
    **kwargs,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Read mercator tile from an image.

    Attributes
    ----------
        src_dst : rasterio.io.DatasetReader
            rasterio.io.DatasetReader object
        x : int
            Mercator tile X index.
        y : int
            Mercator tile Y index.
        z : int
            Mercator tile ZOOM level.
        tilesize : int, optional
            Output tile size. Default is 256.
        kwargs : Any, optional
            Additional options to forward to part()

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    bounds = transform_bounds(
        src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
    )
    if not tile_exists(bounds, z, x, y):
        raise TileOutsideBounds(f"Tile {z}/{x}/{y} is outside image bounds")

    tile_bounds = mercantile.xy_bounds(mercantile.Tile(x=x, y=y, z=z))
    return part(src_dst, tile_bounds, tilesize, tilesize, **kwargs)


def multi_tile(
    assets: Sequence[str], *args: Any, **kwargs: Any
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Assemble multiple rio_tiler.reader.tile."""

    def worker(asset: str):
        with rasterio.open(asset) as src_dst:
            return tile(src_dst, *args, **kwargs)

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        data, masks = zip(*list(executor.map(worker, assets)))
        data = numpy.concatenate(data)
        mask = numpy.all(masks, axis=0).astype(numpy.uint8) * 255
        return data, mask


def multi_preview(
    assets: Sequence[str], *args: Any, **kwargs: Any
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Assemble multiple rio_tiler.reader.preview."""

    def worker(asset: str):
        with rasterio.open(asset) as src_dst:
            return preview(src_dst, *args, **kwargs)

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        data, masks = zip(*list(executor.map(worker, assets)))
        data = numpy.concatenate(data)
        mask = numpy.all(masks, axis=0).astype(numpy.uint8) * 255
        return data, mask


def multi_point(assets: Sequence[str], *args: Any, **kwargs: Any) -> Sequence:
    """Assemble multiple rio_tiler.reader.point."""

    def worker(asset: str):
        with rasterio.open(asset) as src_dst:
            return point(src_dst, *args, **kwargs)

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        return list(executor.map(worker, assets))


def multi_metadata(assets: Sequence[str], *args: Any, **kwargs: Any) -> Sequence:
    """Assemble multiple rio_tiler.reader.metadata."""

    def worker(asset: str):
        with rasterio.open(asset) as src_dst:
            return metadata(src_dst, *args, **kwargs)

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        return list(executor.map(worker, assets))



"""rio_tiler.io.landsat8: Landsat-8 processing."""

from typing import Any, Dict, Sequence, Tuple, Union

import os
import re
import datetime
from concurrent import futures

from urllib.request import urlopen

import numpy

import rasterio
from rasterio.warp import transform_bounds
from rio_toa import reflectance, brightness_temp, toa_utils

from rio_tiler import reader
from rio_tiler import constants
from rio_tiler.utils import _stats as raster_stats, tile_exists, pansharpening_brovey
from rio_tiler.errors import (
    TileOutsideBounds,
    InvalidBandName,
    InvalidLandsatSceneId,
)

LANDSAT_BANDS = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "QA"]


def landsat_parser(sceneid: str) -> Dict:
    """
    Parse Landsat-8 scene id.

    Author @perrygeo - http://www.perrygeo.com

    Attributes
    ----------
        sceneid : str
            Landsat sceneid.

    Returns
    -------
        out : dict
            dictionary with metadata constructed from the sceneid.

    """
    pre_collection = r"(L[COTEM]8\d{6}\d{7}[A-Z]{3}\d{2})"
    collection_1 = r"(L[COTEM]08_L\d{1}[A-Z]{2}_\d{6}_\d{8}_\d{8}_\d{2}_(T1|T2|RT))"
    if not re.match("^{}|{}$".format(pre_collection, collection_1), sceneid):
        raise InvalidLandsatSceneId("Could not match {}".format(sceneid))

    precollection_pattern = (
        r"^L"
        r"(?P<sensor>\w{1})"
        r"(?P<satellite>\w{1})"
        r"(?P<path>[0-9]{3})"
        r"(?P<row>[0-9]{3})"
        r"(?P<acquisitionYear>[0-9]{4})"
        r"(?P<acquisitionJulianDay>[0-9]{3})"
        r"(?P<groundStationIdentifier>\w{3})"
        r"(?P<archiveVersion>[0-9]{2})$"
    )

    collection_pattern = (
        r"^L"
        r"(?P<sensor>\w{1})"
        r"(?P<satellite>\w{2})"
        r"_"
        r"(?P<processingCorrectionLevel>\w{4})"
        r"_"
        r"(?P<path>[0-9]{3})"
        r"(?P<row>[0-9]{3})"
        r"_"
        r"(?P<acquisitionYear>[0-9]{4})"
        r"(?P<acquisitionMonth>[0-9]{2})"
        r"(?P<acquisitionDay>[0-9]{2})"
        r"_"
        r"(?P<processingYear>[0-9]{4})"
        r"(?P<processingMonth>[0-9]{2})"
        r"(?P<processingDay>[0-9]{2})"
        r"_"
        r"(?P<collectionNumber>\w{2})"
        r"_"
        r"(?P<collectionCategory>\w{2})$"
    )

    for pattern in [collection_pattern, precollection_pattern]:
        match = re.match(pattern, sceneid, re.IGNORECASE)
        if match:
            meta: Dict[str, Any] = match.groupdict()
            break

    meta["scene"] = sceneid
    if meta.get("acquisitionJulianDay"):
        date = datetime.datetime(
            int(meta["acquisitionYear"]), 1, 1
        ) + datetime.timedelta(int(meta["acquisitionJulianDay"]) - 1)

        meta["date"] = date.strftime("%Y-%m-%d")
    else:
        meta["date"] = "{}-{}-{}".format(
            meta["acquisitionYear"], meta["acquisitionMonth"], meta["acquisitionDay"]
        )

    collection = meta.get("collectionNumber", "")
    if collection != "":
        collection = "c{}".format(int(collection))

    meta["scheme"] = "s3"
    meta["bucket"] = "landsat-pds"
    meta["prefix"] = os.path.join(collection, "L8", meta["path"], meta["row"], sceneid)

    return meta


def _landsat_get_mtl(sceneid: str) -> Dict:
    """
    Get Landsat-8 MTL metadata.

    Attributes
    ----------
        sceneid : str
            Landsat sceneid. For scenes after May 2017,
            sceneid have to be LANDSAT_PRODUCT_ID.

    Returns
    -------
        out : dict
            returns a JSON like object with the metadata.

    """
    scene_params = landsat_parser(sceneid)
    meta_file = "http://{bucket}.s3.amazonaws.com/{prefix}/{scene}_MTL.txt".format(
        **scene_params
    )
    metadata = str(urlopen(meta_file).read().decode())
    return toa_utils._parse_mtl_txt(metadata)


def _convert(arr: numpy.ndarray, band: str, metadata: Dict) -> numpy.ndarray:
    """Convert DN to TOA or Temp."""
    if band in ["1", "2", "3", "4", "5", "6", "7", "8", "9"]:  # OLI
        multi_reflect = metadata["RADIOMETRIC_RESCALING"].get(
            f"REFLECTANCE_MULT_BAND_{band}"
        )
        add_reflect = metadata["RADIOMETRIC_RESCALING"].get(
            f"REFLECTANCE_ADD_BAND_{band}"
        )
        sun_elev = metadata["IMAGE_ATTRIBUTES"]["SUN_ELEVATION"]

        arr = 10000 * reflectance.reflectance(
            arr, multi_reflect, add_reflect, sun_elev, src_nodata=0
        )

    elif band in ["10", "11"]:  # TIRS
        multi_rad = metadata["RADIOMETRIC_RESCALING"].get(f"RADIANCE_MULT_BAND_{band}")
        add_rad = metadata["RADIOMETRIC_RESCALING"].get(f"RADIANCE_ADD_BAND_{band}")
        k1 = metadata["TIRS_THERMAL_CONSTANTS"].get(f"K1_CONSTANT_BAND_{band}")
        k2 = metadata["TIRS_THERMAL_CONSTANTS"].get(f"K2_CONSTANT_BAND_{band}")

        arr = brightness_temp.brightness_temp(arr, multi_rad, add_rad, k1, k2)

    # TODO
    # elif band == "QA":

    return arr


def bounds(sceneid: str) -> Dict:
    """
    Retrieve image bounds.

    Attributes
    ----------
        sceneid : str
            Landsat sceneid. For scenes after May 2017,
            sceneid have to be LANDSAT_PRODUCT_ID.

    Returns
    -------
        out : dict
            dictionary with image bounds.

    """
    meta: Dict = _landsat_get_mtl(sceneid)["L1_METADATA_FILE"]

    return dict(
        sceneid=sceneid,
        bounds=toa_utils._get_bounds_from_metadata(meta["PRODUCT_METADATA"]),
    )


def metadata(
    sceneid: str,
    pmin: float = 2.0,
    pmax: float = 98.0,
    hist_options: Dict = {},
    **kwargs: Any,
) -> Dict:
    """
    Retrieve image bounds and band statistics.

    Attributes
    ----------
        sceneid : str
            Landsat sceneid. For scenes after May 2017,
            sceneid have to be LANDSAT_PRODUCT_ID.
        pmin : int, optional, (default: 2)
            Histogram minimum cut.
        pmax : int, optional, (default: 98)
            Histogram maximum cut.
        hist_options : dict, optional
            Options to forward to numpy.histogram function.
            e.g: {bins=20, range=(0, 1000)}
        kwargs : optional
            These are passed to 'rio_tiler.reader.preview'

    Returns
    -------
        out : dict
            Dictionary with bounds and bands statistics.

    """
    scene_params = landsat_parser(sceneid)
    meta: Dict = _landsat_get_mtl(sceneid)["L1_METADATA_FILE"]

    landsat_prefix = "{scheme}://{bucket}/{prefix}/{scene}".format(**scene_params)

    def worker(band: str):
        asset = f"{landsat_prefix}_B{band}.TIF"

        if band == "QA":
            nodata = 1
            resamp = "nearest"
        else:
            nodata = 0
            resamp = "bilinear"

        with rasterio.open(asset) as src_dst:
            bounds = transform_bounds(
                src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
            )
            data, mask = reader.preview(
                src_dst, nodata=nodata, resampling_method=resamp, **kwargs
            )

        if band != "QA":
            data = data.astype("float32", casting="unsafe")
            data = _convert(data, band, meta)

        data = numpy.ma.array(data)
        data.mask = mask == 0

        statistics = raster_stats(data, percentiles=(pmin, pmax), **hist_options)
        return dict(bounds=bounds, statistics=statistics)

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        responses = list(executor.map(worker, LANDSAT_BANDS))

    info: Dict[str, Any] = dict(sceneid=sceneid)
    info["band_descriptions"] = [(ix + 1, b) for ix, b in enumerate(LANDSAT_BANDS)]
    info["bounds"] = [
        r["bounds"] for b, r in zip(LANDSAT_BANDS, responses) if b == "8"
    ][0]

    info["statistics"] = {b: d["statistics"] for b, d in zip(LANDSAT_BANDS, responses)}
    return info


def tile(
    sceneid: str,
    tile_x: int,
    tile_y: int,
    tile_z: int,
    bands: Union[Sequence[str], str] = ["4", "3", "2"],
    tilesize: int = 256,
    pan: bool = False,
    **kwargs: Any,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Create mercator tile from Landsat-8 data.

    Attributes
    ----------
        sceneid : str
            Landsat sceneid. For scenes after May 2017,
            sceneid have to be LANDSAT_PRODUCT_ID.
        tile_x : int
            Mercator tile X index.
        tile_y : int
            Mercator tile Y index.
        tile_z : int
            Mercator tile ZOOM level.
        bands : tuple, str, optional (default: ("4", "3", "2"))
            Bands index for the RGB combination.
        tilesize : int, optional (default: 256)
            Output image size.
        pan : boolean, optional (default: False)
            If True, apply pan-sharpening.
        kwargs: dict, optional
            These will be passed to the 'rio_tiler.utils._tile_read' function.

    Returns
    -------
    data : numpy ndarray
    mask: numpy array

    """
    if isinstance(bands, str):
        bands = (bands,)

    for band in bands:
        if band not in LANDSAT_BANDS:
            raise InvalidBandName("{} is not a valid Landsat band name".format(band))

    scene_params = landsat_parser(sceneid)

    meta: Dict = _landsat_get_mtl(sceneid)["L1_METADATA_FILE"]

    landsat_prefix = "{scheme}://{bucket}/{prefix}/{scene}".format(**scene_params)

    bounds = toa_utils._get_bounds_from_metadata(meta["PRODUCT_METADATA"])
    if not tile_exists(bounds, tile_z, tile_x, tile_y):
        raise TileOutsideBounds(
            "Tile {}/{}/{} is outside image bounds".format(tile_z, tile_x, tile_y)
        )

    def worker(band: str):
        asset = f"{landsat_prefix}_B{band}.TIF"

        if band == "QA":
            nodata = 1
            resamp = "nearest"
        else:
            nodata = 0
            resamp = "bilinear"

        with rasterio.open(asset) as src_dst:
            tile, mask = reader.tile(
                src_dst,
                tile_x,
                tile_y,
                tile_z,
                tilesize=tilesize,
                nodata=nodata,
                resampling_method=resamp,
            )

        return tile, mask

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        data, masks = zip(*list(executor.map(worker, bands)))
        data = numpy.concatenate(data)
        mask = numpy.all(masks, axis=0).astype(numpy.uint8) * 255

        if pan:
            pan_data, mask = worker("8")
            data = pansharpening_brovey(data, pan_data, 0.2, pan_data.dtype)

        if bands[0] != "QA" or len(bands) != 1:
            for bdx, band in enumerate(bands):
                data[bdx] = _convert(data[bdx], band, meta)

        return data, mask



"""rio_tiler.io.cogeo: raster processing."""

from typing import Any, Dict, Tuple

import numpy

import rasterio
from rasterio.warp import transform_bounds

from rio_tiler import reader
from rio_tiler import constants
from rio_tiler.utils import has_alpha_band, has_mask_band
from rio_tiler.mercator import get_zooms


def spatial_info(address: str) -> Dict:
    """
    Return COGEO spatial info.

    Attributes
    ----------
        address : str or PathLike object
            A dataset path or URL. Will be opened in "r" mode.

    Returns
    -------
        out : dict.

    """
    with rasterio.open(address) as src_dst:
        minzoom, maxzoom = get_zooms(src_dst)
        bounds = transform_bounds(
            src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
        )
        center = [(bounds[0] + bounds[2]) / 2, (bounds[1] + bounds[3]) / 2, minzoom]

    return dict(
        address=address, bounds=bounds, center=center, minzoom=minzoom, maxzoom=maxzoom
    )


def bounds(address: str) -> Dict:
    """
    Retrieve image bounds.

    Attributes
    ----------
        address : str
            file url.

    Returns
    -------
        out : dict
            dictionary with image bounds.

    """
    with rasterio.open(address) as src_dst:
        bounds = transform_bounds(
            src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
        )
    return dict(address=address, bounds=bounds)


def metadata(
    address: str,
    pmin: float = 2.0,
    pmax: float = 98.0,
    hist_options: Dict = {},
    **kwargs: Any,
) -> Dict:
    """
    Return image statistics.

    Attributes
    ----------
        address : str or PathLike object
            A dataset path or URL. Will be opened in "r" mode.
        pmin : int, optional, (default: 2)
            Histogram minimum cut.
        pmax : int, optional, (default: 98)
            Histogram maximum cut.
        hist_options : dict, optional
            Options to forward to numpy.histogram function.
            e.g: {bins=20, range=(0, 1000)}
        kwargs : optional
            These are passed to 'rio_tiler.reader.preview'

    Returns
    -------
        out : dict
            Dictionary with image bounds and bands statistics.

    """
    with rasterio.open(address) as src_dst:
        meta = reader.metadata(
            src_dst, percentiles=(pmin, pmax), hist_options=hist_options, **kwargs
        )

    return dict(address=address, **meta)


def info(address: str) -> Dict:
    """
    Return simple metadata about the file.

    Attributes
    ----------
    address : str or PathLike object
        A dataset path or URL. Will be opened in "r" mode.

    Returns
    -------
    out : dict.

    """
    with rasterio.open(address) as src_dst:
        minzoom, maxzoom = get_zooms(src_dst)
        bounds = transform_bounds(
            src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
        )
        center = [(bounds[0] + bounds[2]) / 2, (bounds[1] + bounds[3]) / 2, minzoom]

        def _get_descr(ix):
            """Return band description."""
            name = src_dst.descriptions[ix - 1]
            if not name:
                name = "band{}".format(ix)
            return name

        band_descriptions = [(ix, _get_descr(ix)) for ix in src_dst.indexes]
        tags = [(ix, src_dst.tags(ix)) for ix in src_dst.indexes]

        other_meta = dict()
        if src_dst.scales[0] and src_dst.offsets[0]:
            other_meta.update(dict(scale=src_dst.scales[0]))
            other_meta.update(dict(offset=src_dst.offsets[0]))

        if has_alpha_band(src_dst):
            nodata_type = "Alpha"
        elif has_mask_band(src_dst):
            nodata_type = "Mask"
        elif src_dst.nodata is not None:
            nodata_type = "Nodata"
        else:
            nodata_type = "None"

        try:
            cmap = src_dst.colormap(1)
            other_meta.update(dict(colormap=cmap))
        except ValueError:
            pass

        return dict(
            address=address,
            bounds=bounds,
            center=center,
            minzoom=minzoom,
            maxzoom=maxzoom,
            band_metadata=tags,
            band_descriptions=band_descriptions,
            dtype=src_dst.meta["dtype"],
            colorinterp=[src_dst.colorinterp[ix - 1].name for ix in src_dst.indexes],
            nodata_type=nodata_type,
            **other_meta,
        )


def tile(
    address: str,
    tile_x: int,
    tile_y: int,
    tile_z: int,
    tilesize: int = 256,
    **kwargs: Any,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Create mercator tile from any images.

    Attributes
    ----------
        address : str
            file url.
        tile_x : int
            Mercator tile X index.
        tile_y : int
            Mercator tile Y index.
        tile_z : int
            Mercator tile ZOOM level.
        tilesize : int, optional (default: 256)
            Output image size.
        kwargs: dict, optional
            These will be passed to the 'rio_tiler.reader.tile' function.

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    with rasterio.open(address) as src_dst:
        return reader.tile(src_dst, tile_x, tile_y, tile_z, tilesize, **kwargs)



"""rio_tiler.io.stac: STAC reader."""

from typing import Any, Dict, Sequence, Tuple

import numpy

from rio_tiler import reader
from rio_tiler.utils import tile_exists
from rio_tiler.errors import InvalidBandName, TileOutsideBounds


def _get_href(stac: Dict, assets: Sequence[str]) -> Sequence[str]:
    """Validate asset names and return asset's url."""
    _assets = list(stac["assets"].keys())
    for asset in assets:
        if asset not in _assets:
            raise InvalidBandName(f"{asset} is not a valid asset name.")

    return [stac["assets"][asset]["href"] for asset in assets]


def spatial_info(stac: Dict) -> Dict:
    """
    Return STAC spatial info.

    Attributes
    ----------
        stac : dict
            STAC item.

    Returns
    -------
        out : dict.

    """
    raise Exception("Not implemented")


def bounds(stac: Dict) -> Dict:
    """
    Return STAC bounds.

    Attributes
    ----------
        stac : dict
            STAC item.

    Returns
    -------
        out : dict
            dictionary with image bounds.

    """
    return dict(id=stac["id"], bounds=stac["bbox"])


def metadata(
    stac: Dict,
    assets: Sequence[str],
    pmin: float = 2.0,
    pmax: float = 98.0,
    hist_options: Dict = {},
    **kwargs: Any,
) -> Dict:
    """
    Return STAC assets statistics.

    Attributes
    ----------
        stac : dict
            STAC item.
        assets : list
            Asset names.
        pmin : int, optional, (default: 2)
            Histogram minimum cut.
        pmax : int, optional, (default: 98)
            Histogram maximum cut.
        hist_options : dict, optional
            Options to forward to numpy.histogram function.
            e.g: {bins=20, range=(0, 1000)}
        kwargs : optional
            These are passed to 'rio_tiler.reader.preview'

    Returns
    -------
        out : dict
            Dictionary with image bounds and bands statistics.

    """
    if isinstance(assets, str):
        assets = (assets,)

    assets_url = _get_href(stac, assets)
    responses = reader.multi_metadata(
        assets_url, percentiles=(pmin, pmax), hist_options=hist_options, **kwargs
    )

    info: Dict[str, Any] = dict(id=stac["id"])
    info["band_descriptions"] = [(ix + 1, b) for ix, b in enumerate(assets)]
    info["bounds"] = stac["bbox"]
    info["statistics"] = {b: d["statistics"][1] for b, d in zip(assets, responses)}
    info["dtypes"] = {b: d["dtype"] for b, d in zip(assets, responses)}
    info["nodata_types"] = {b: d["nodata_type"] for b, d in zip(assets, responses)}
    return info


def tile(
    stac: Dict,
    assets: Sequence[str],
    tile_x: int,
    tile_y: int,
    tile_z: int,
    tilesize: int = 256,
    **kwargs: Any,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Create mercator tile from any images.

    Attributes
    ----------
        stac : dict
            STAC item.
        assets : list
            Asset names.
        tile_x : int
            Mercator tile X index.
        tile_y : int
            Mercator tile Y index.
        tile_z : int
            Mercator tile ZOOM level.
        tilesize : int, optional (default: 256)
            Output image size.
        kwargs: dict, optional
            These will be passed to the 'rio_tiler.reader.tile' function.

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    if isinstance(assets, str):
        assets = (assets,)

    if not tile_exists(stac["bbox"], tile_z, tile_x, tile_y):
        raise TileOutsideBounds(
            f"Tile {tile_z}/{tile_x}/{tile_y} is outside item bounds"
        )

    assets_url = _get_href(stac, assets)
    return reader.multi_tile(assets_url, tile_x, tile_y, tile_z, **kwargs)



"""rio_tiler.cbers: cbers processing."""

from typing import Any, Dict, Sequence, Tuple, Union

import re

import numpy

import rasterio
from rasterio.warp import transform_bounds

from rio_tiler import reader
from rio_tiler import constants
from rio_tiler.utils import tile_exists
from rio_tiler.errors import TileOutsideBounds, InvalidBandName, InvalidCBERSSceneId


def cbers_parser(sceneid: str) -> Dict:
    """Parse CBERS scene id.

    Attributes
    ----------
        sceneid : str
            CBERS sceneid.

    Returns
    -------
        out : dict
            dictionary with metadata constructed from the sceneid.

    """
    if not re.match(r"^CBERS_4_\w+_[0-9]{8}_[0-9]{3}_[0-9]{3}_L[0-9]$", sceneid):
        raise InvalidCBERSSceneId("Could not match {}".format(sceneid))

    cbers_pattern = (
        r"(?P<satellite>\w+)_"
        r"(?P<mission>[0-9]{1})"
        r"_"
        r"(?P<instrument>\w+)"
        r"_"
        r"(?P<acquisitionYear>[0-9]{4})"
        r"(?P<acquisitionMonth>[0-9]{2})"
        r"(?P<acquisitionDay>[0-9]{2})"
        r"_"
        r"(?P<path>[0-9]{3})"
        r"_"
        r"(?P<row>[0-9]{3})"
        r"_"
        r"(?P<processingCorrectionLevel>L[0-9]{1})$"
    )

    meta: Dict[str, Any] = re.match(cbers_pattern, sceneid, re.IGNORECASE).groupdict()
    meta["scene"] = sceneid

    instrument = meta["instrument"]
    instrument_params = {
        "MUX": {
            "reference_band": "6",
            "bands": ("5", "6", "7", "8"),
            "rgb": ("7", "6", "5"),
        },
        "AWFI": {
            "reference_band": "14",
            "bands": ("13", "14", "15", "16"),
            "rgb": ("15", "14", "13"),
        },
        "PAN10M": {
            "reference_band": "4",
            "bands": ("2", "3", "4"),
            "rgb": ("3", "4", "2"),
        },
        "PAN5M": {"reference_band": "1", "bands": ("1"), "rgb": ("1", "1", "1")},
    }
    meta["reference_band"] = instrument_params[instrument]["reference_band"]
    meta["bands"] = instrument_params[instrument]["bands"]
    meta["rgb"] = instrument_params[instrument]["rgb"]

    meta["scheme"] = "s3"
    meta["bucket"] = "cbers-pds"
    meta["prefix"] = "CBERS4/{instrument}/{path}/{row}/{scene}".format(**meta)

    return meta


def bounds(sceneid: str) -> Dict:
    """
    Retrieve image bounds.

    Attributes
    ----------
        sceneid : str
            CBERS sceneid.

    Returns
    -------
        out : dict
            dictionary with image bounds.

    """
    scene_params = cbers_parser(sceneid)
    cbers_prefix = "{scheme}://{bucket}/{prefix}/{scene}".format(**scene_params)

    with rasterio.open(
        "{}_BAND{}.tif".format(cbers_prefix, scene_params["reference_band"])
    ) as src:
        bounds = transform_bounds(
            src.crs, constants.WGS84_CRS, *src.bounds, densify_pts=21
        )

    return dict(sceneid=sceneid, bounds=bounds)


def metadata(
    sceneid: str,
    pmin: float = 2.0,
    pmax: float = 98.0,
    hist_options: Dict = {},
    **kwargs: Any,
) -> Dict:
    """
    Return band bounds and statistics.

    Attributes
    ----------
        sceneid : str
            CBERS sceneid.
        pmin : int, optional, (default: 2)
            Histogram minimum cut.
        pmax : int, optional, (default: 98)
            Histogram maximum cut.
        hist_options : dict, optional
            Options to forward to numpy.histogram function.
            e.g: {bins=20, range=(0, 1000)}
        kwargs : optional
            These are passed to 'rio_tiler.reader.preview'

    Returns
    -------
        out : dict
            Dictionary with bounds and bands statistics.

    """
    scene_params = cbers_parser(sceneid)
    cbers_prefix = "{scheme}://{bucket}/{prefix}/{scene}".format(**scene_params)

    bands = scene_params["bands"]
    addresses = [f"{cbers_prefix}_BAND{band}.tif" for band in bands]

    responses = reader.multi_metadata(
        addresses,
        indexes=[1],
        nodata=0,
        percentiles=(pmin, pmax),
        hist_options=hist_options,
        **kwargs,
    )

    info: Dict[str, Any] = dict(sceneid=sceneid)
    info["instrument"] = scene_params["instrument"]
    info["band_descriptions"] = [(ix + 1, b) for ix, b in enumerate(bands)]

    info["bounds"] = [
        r["bounds"]
        for b, r in zip(bands, responses)
        if b == scene_params["reference_band"]
    ][0]
    info["statistics"] = {b: d["statistics"][1] for b, d in zip(bands, responses)}
    return info


def tile(
    sceneid: str,
    tile_x: int,
    tile_y: int,
    tile_z: int,
    bands: Union[Sequence[str], str] = None,
    tilesize: int = 256,
    **kwargs: Dict,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Create mercator tile from CBERS data.

    Attributes
    ----------
        sceneid : str
            CBERS sceneid.
        tile_x : int
            Mercator tile X index.
        tile_y : int
            Mercator tile Y index.
        tile_z : int
            Mercator tile ZOOM level.
        bands : tuple or list or str, optional
            Bands index for the RGB combination. If None uses default
            defined for the instrument
        tilesize : int, optional
            Output image size. Default is 256
        kwargs: dict, optional
            These will be passed to the 'rio_tiler.reader.tile' function.

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    if isinstance(bands, str):
        bands = (bands,)

    scene_params = cbers_parser(sceneid)

    if not bands:
        bands = scene_params["rgb"]

    for band in bands:
        if band not in scene_params["bands"]:
            raise InvalidBandName(
                "{} is not a valid band name for {} CBERS instrument".format(
                    band, scene_params["instrument"]
                )
            )

    cbers_prefix = "{scheme}://{bucket}/{prefix}/{scene}".format(**scene_params)
    with rasterio.open(
        "{}_BAND{}.tif".format(cbers_prefix, scene_params["reference_band"])
    ) as src_dst:
        bounds = transform_bounds(
            src_dst.crs, constants.WGS84_CRS, *src_dst.bounds, densify_pts=21
        )

    if not tile_exists(bounds, tile_z, tile_x, tile_y):
        raise TileOutsideBounds(
            "Tile {}/{}/{} is outside image bounds".format(tile_z, tile_x, tile_y)
        )

    addresses = [f"{cbers_prefix}_BAND{band}.tif" for band in bands]
    return reader.multi_tile(
        addresses, tile_x, tile_y, tile_z, tilesize=tilesize, nodata=0
    )



"""Setup for rio-tiler."""

from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()

# Runtime requirements.
inst_reqs = ["numpy", "numexpr", "mercantile", "rasterio[s3]>=1.1.3", "rio-toa"]

extra_reqs = {
    "test": ["mock", "pytest", "pytest-benchmark", "pytest-cov", "rio-cogeo"],
    "dev": [
        "mock",
        "pytest",
        "pytest-benchmark",
        "pytest-cov",
        "rio-cogeo",
        "pre-commit",
    ],
}

setup(
    name="rio-tiler",
    version="2.0a5",
    python_requires=">=3",
    description=u"""Get mercator tile from CloudOptimized GeoTIFF and other cloud hosted raster such as CBERS-4, Sentinel-2, Sentinel-1 and Landsat-8 AWS PDS""",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    keywords="COG cogeo raster aws map tiler gdal rasterio",
    author=u"Vincent Sarago",
    author_email="vincent@developmentseed.org",
    url="https://github.com/cogeotiff/rio-tiler",
    license="BSD",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    include_package_data=True,
    package_data={"rio_tiler": ["cmap/*.npy"]},
    zip_safe=False,
    install_requires=inst_reqs,
    extras_require=extra_reqs,
)



"""rio_tiler.io.sentinel1: Sentinel-1 processing."""

from typing import Any, Dict, Sequence, Tuple, Union

import os
import re
import json
from concurrent import futures

import numpy

from boto3.session import Session as boto3_session

import rasterio
from rasterio import transform
from rasterio.vrt import WarpedVRT

from rio_tiler import reader
from rio_tiler import constants
from rio_tiler.errors import InvalidBandName, InvalidSentinelSceneId

REGION = os.environ.get("AWS_REGION", "eu-central-1")
SENTINEL_BANDS = ["vv", "vh"]


def _aws_get_object(
    bucket: str,
    key: str,
    request_pays: bool = True,
    client: boto3_session.client = None,
) -> bytes:
    """AWS s3 get object content."""
    if not client:
        session = boto3_session(region_name=REGION)
        client = session.client("s3")

    params = {"Bucket": bucket, "Key": key}
    if request_pays:
        params["RequestPayer"] = "requester"

    response = client.get_object(**params)

    return response["Body"].read()


def sentinel1_parser(sceneid: str) -> Dict:
    """
    Parse Sentinel-1 scene id.

    Attributes
    ----------
        sceneid : str
            Sentinel-1 sceneid.

    Returns
    -------
        out : dict
            dictionary with metadata constructed from the sceneid.

    """
    if not re.match(
        "^S1[AB]_(IW)|(EW)_[A-Z]{3}[FHM]_[0-9][SA][A-Z]{2}_[0-9]{8}T[0-9]{6}_[0-9]{8}T[0-9]{6}_[0-9A-Z]{6}_[0-9A-Z]{6}_[0-9A-Z]{4}$",
        sceneid,
    ):
        raise InvalidSentinelSceneId("Could not match {}".format(sceneid))

    sentinel_pattern = (
        r"^S"
        r"(?P<sensor>\w{1})"
        r"(?P<satellite>[AB]{1})"
        r"_"
        r"(?P<beam>[A-Z]{2})"
        r"_"
        r"(?P<product>[A-Z]{3})"
        r"(?P<resolution>[FHM])"
        r"_"
        r"(?P<processing_level>[0-9])"
        r"(?P<product_class>[SA])"
        r"(?P<polarisation>(SH)|(SV)|(DH)|(DV)|(HH)|(HV)|(VV)|(VH))"
        r"_"
        r"(?P<startDateTime>[0-9]{8}T[0-9]{6})"
        r"_"
        r"(?P<stopDateTime>[0-9]{8}T[0-9]{6})"
        r"_"
        r"(?P<absolute_orbit>[0-9]{6})"
        r"_"
        r"(?P<mission_task>[0-9A-Z]{6})"
        r"_"
        r"(?P<product_id>[0-9A-Z]{4})$"
    )

    meta: Dict[str, Any] = re.match(
        sentinel_pattern, sceneid, re.IGNORECASE
    ).groupdict()

    meta["scene"] = sceneid
    year = meta["startDateTime"][0:4]
    month = meta["startDateTime"][4:6].strip("0")
    day = meta["startDateTime"][6:8].strip("0")

    meta["scheme"] = "s3"
    meta["bucket"] = "sentinel-s1-l1c"
    meta["prefix"] = os.path.join(
        meta["product"], year, month, day, meta["beam"], meta["polarisation"], sceneid
    )

    return meta


def _get_bounds(scene_info: Dict) -> Tuple[float, float, float, float]:
    bucket, prefix = scene_info["bucket"], scene_info["prefix"]
    product_info = json.loads(_aws_get_object(bucket, f"{prefix}/productInfo.json"))

    xyz = list(zip(*product_info["footprint"]["coordinates"][0]))
    return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])


def bounds(sceneid: str) -> Dict:
    """
    Retrieve image bounds.

    Attributes
    ----------
        sceneid : str
            Sentinel-1 sceneid.

    Returns
    -------
        out : dict
            dictionary with image bounds.

    """
    scene_params = sentinel1_parser(sceneid)
    return dict(sceneid=sceneid, bounds=_get_bounds(scene_params))


def metadata(
    sceneid: str,
    pmin: float = 2.0,
    pmax: float = 98.0,
    bands: Union[Sequence[str], str] = None,
    hist_options: Dict = {},
    **kwargs,
) -> Dict:
    """
    Retrieve image bounds and band statistics.

    Attributes
    ----------
        sceneid : str
            Sentinel-1 sceneid.
        pmin : float, optional, (default: 2.)
            Histogram minimum cut.
        pmax : float, optional, (default: 98.)
            Histogram maximum cut.
        bands: tuple, str, required
            Bands name (e.g vv, vh).
        kwargs : optional
            These are passed to 'rio_tiler.utils._stats'
            e.g: bins=20, range=(0, 1000)

    Returns
    -------
        out : dict
            Dictionary with image bounds and bands statistics.

    """
    if not bands:
        raise InvalidBandName("bands is required")

    if isinstance(bands, str):
        bands = (bands,)

    for band in bands:
        if band not in SENTINEL_BANDS:
            raise InvalidBandName("{} is not a valid Sentinel band name".format(band))

    scene_params = sentinel1_parser(sceneid)
    sentinel_prefix = "{scheme}://{bucket}/{prefix}/measurement".format(**scene_params)

    def worker(band: str):
        asset = "{}/{}-{}.tiff".format(
            sentinel_prefix, scene_params["beam"].lower(), band
        )
        with rasterio.open(asset) as src_dst:
            with WarpedVRT(
                src_dst,
                src_crs=src_dst.gcps[1],
                src_transform=transform.from_gcps(src_dst.gcps[0]),
                src_nodata=0,
            ) as vrt_dst:
                return reader.metadata(
                    vrt_dst,
                    percentiles=(pmin, pmax),
                    hist_options=hist_options,
                    **kwargs,
                )

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        responses = list(executor.map(worker, bands))

    info = dict(
        sceneid=sceneid,
        bounds=responses[0]["bounds"],
        band_descriptions=[(ix + 1, b) for ix, b in enumerate(bands)],
    )

    info["statistics"] = {
        b: v for b, d in zip(bands, responses) for _, v in d["statistics"].items()
    }
    return info


def tile(
    sceneid: str,
    tile_x: int,
    tile_y: int,
    tile_z: int,
    bands: Union[Sequence[str], str] = None,
    tilesize: int = 256,
    **kwargs: Any,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Create mercator tile from Sentinel-1 data.

    Attributes
    ----------
        sceneid : str
            Sentinel-2 sceneid.
        tile_x : int
            Mercator tile X index.
        tile_y : int
            Mercator tile Y index.
        tile_z : int
            Mercator tile ZOOM level.
        bands: tuple, str, required
            Bands name (e.g vv, vh).
        tilesize : int, optional (default: 256)
            Output image size.

    Returns
    -------
        data : numpy ndarray
        mask: numpy array

    """
    if not bands:
        raise InvalidBandName("bands is required")

    if isinstance(bands, str):
        bands = (bands,)

    for band in bands:
        if band not in SENTINEL_BANDS:
            raise InvalidBandName("{} is not a valid Sentinel band name".format(band))

    scene_params = sentinel1_parser(sceneid)
    sentinel_prefix = "{scheme}://{bucket}/{prefix}/measurement".format(**scene_params)

    def worker(band: str):
        asset = "{}/{}-{}.tiff".format(
            sentinel_prefix, scene_params["beam"].lower(), band
        )
        with rasterio.open(asset) as src_dst:
            with WarpedVRT(
                src_dst,
                src_crs=src_dst.gcps[1],
                src_transform=transform.from_gcps(src_dst.gcps[0]),
                src_nodata=0,
            ) as vrt_dst:
                return reader.tile(
                    vrt_dst, tile_x, tile_y, tile_z, tilesize=tilesize, **kwargs
                )

    with futures.ThreadPoolExecutor(max_workers=constants.MAX_THREADS) as executor:
        data, masks = zip(*list(executor.map(worker, bands)))
        data = numpy.concatenate(data)
        mask = numpy.all(masks, axis=0).astype(numpy.uint8) * 255

    return data, mask



"""rio-tiler colormap functions."""

from typing import Dict, Sequence, Tuple

import os

import numpy


EMPTY_COLORMAP: Dict = {i: [0, 0, 0, 0] for i in range(256)}


def _update_alpha(cmap: Dict, idx: Sequence[int], alpha: int = 0) -> None:
    """Update the alpha value of a colormap index."""
    if isinstance(idx, int):
        idx = (idx,)
    for i in idx:
        cmap[i] = cmap[i][0:3] + [alpha]


def _remove_value(cmap: Dict, idx: Sequence[int]) -> None:
    """Remove value from a colormap dict."""
    if isinstance(idx, int):
        idx = (idx,)

    for i in idx:
        cmap.pop(i, None)


def _update_cmap(cmap: Dict, values: Dict) -> None:
    """Update a colormap dict."""
    for i, color in values.items():
        if len(color) == 3:
            color += [255]
        cmap[i] = color


def get_colormap(name: str) -> Dict:
    """
    Return colormap dict.

    Attributes
    ----------
        name : str, optional
            Colormap name (default: cfastie)

    Returns
    -------
        colormap : dict
            GDAL RGBA Color Table dictionary.

    """
    cmap_file = os.path.join(os.path.dirname(__file__), "cmap", f"{name.lower()}.npy")
    cmap = numpy.load(cmap_file)
    assert cmap.shape == (256, 4)
    assert cmap.dtype == numpy.uint8

    return {idx: value.tolist() for idx, value in enumerate(cmap)}


# From https://github.com/mojodna/marblecutter/blob/5b9040ba6c83562a465eabdbb6e8959e6a8bf041/marblecutter/utils.py#L35
def make_lut(colormap: Dict) -> numpy.ndarray:
    """
    Create a lookup table numpy.ndarray from a GDAL RGBA Color Table dictionary.

    Attributes
    ----------
        colormap : dict
            GDAL RGBA Color Table dictionary.

    Returns
    -------
        lut : numpy.ndarray
            colormap lookup table

    """
    lut = numpy.zeros(shape=(256, 4), dtype=numpy.uint8)
    for i, color in colormap.items():
        lut[int(i)] = color

    return lut


def apply_cmap(
    data: numpy.ndarray, colormap: Dict
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Apply colormap on tile data.

    Attributes
    ----------
        data : numpy ndarray
            1D image array to translate to RGB.
        colormap : dict
            GDAL RGBA Color Table dictionary.

    Returns
    -------
        data : numpy.ndarray
            RGB data.
        mask: numpy.ndarray
            Alpha band.

    """
    if data.shape[0] > 1:
        raise Exception("Source data must be 1 band")

    lookup_table = make_lut(colormap)
    data = lookup_table[data[0], :]

    data = numpy.transpose(data, [2, 0, 1])

    return data[:-1], data[-1]


def apply_discrete_cmap(
    data: numpy.ndarray, colormap: Dict
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Apply discrete colormap.

    Note: This method is not used by default and left
          to users to use within custom render methods.

    Attributes
    ----------
        data : numpy ndarray
            1D image array to translate to RGB.
        color_map: dict
            Discrete ColorMap dictionary
            e.g:
            {
                1: [255, 255, 255],
                2: [255, 0, 0]
            }
    Returns
    -------
        arr: numpy.ndarray
    """
    res = numpy.zeros((data.shape[1], data.shape[2], 4), dtype=numpy.uint8)

    for k, v in colormap.items():
        res[data[0] == k] = v

    data = numpy.transpose(res, [2, 0, 1])

    return data[:-1], data[-1]



"""Errors and warnings."""


class RioTilerError(Exception):
    """Base exception class."""


class InvalidFormat(RioTilerError):
    """Invalid image format."""


class TileOutsideBounds(RioTilerError):
    """Z-X-Y Tile is outside image bounds."""


class InvalidLandsatSceneId(RioTilerError):
    """Invalid Landsat-8 scene id."""


class InvalidSentinelSceneId(RioTilerError):
    """Invalid Sentinel-2 scene id."""


class InvalidCBERSSceneId(RioTilerError):
    """Invalid CBERS scene id."""


class InvalidBandName(RioTilerError):
    """Invalid band name."""


class DeprecationWarning(UserWarning):
    """Rio-tiler module deprecations warning."""


class AlphaBandWarning(UserWarning):
    """Automaticaly removed Alpha band from output array."""


class NoOverviewWarning(UserWarning):
    """Dataset has no overviews."""
